# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections

import proto
from google.protobuf import descriptor_pb2

from proto import _file_info, _package_info


def sample_file_info(name):
    filename = name + ".proto"

    # Get the essential information about the proto package, and where
    # this component belongs within the file.
    package, marshal = _package_info.compile(name, {})

    # Get or create the information about the file, including the
    # descriptor to which the new message descriptor shall be added.
    return _file_info._FileInfo.registry.setdefault(
        filename,
        _file_info._FileInfo(
            descriptor=descriptor_pb2.FileDescriptorProto(
                name=filename, package=package, syntax="proto3",
            ),
            enums=collections.OrderedDict(),
            messages=collections.OrderedDict(),
            name=filename,
            nested={},
            nested_enum={},
        ),
    )


def test_fallback_salt_is_appended_to_filename():
    # given
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    name = "my-fileinfo"
    fallback_salt = "my-fallback_salt"
    file_info = sample_file_info(name)

    # when
    file_info.generate_file_pb(new_class=Foo, fallback_salt=fallback_salt)

    # then
    assert file_info.descriptor.name == name + "_" + fallback_salt + ".proto"


def test_none_fallback_salt_is_appended_to_filename_as_empty():
    # given

    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    name = "my-fileinfo"
    none_fallback_salt = None
    file_info = sample_file_info(name)

    # when
    file_info.generate_file_pb(new_class=Foo, fallback_salt=none_fallback_salt)

    # then
    assert file_info.descriptor.name == name + ".proto"



# Copyright (C) 2021  Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from proto.primitives import ProtoType


class StringyNumberRule:
    """A marshal between certain numeric types and strings

    This is a necessary hack to allow round trip conversion
    from messages to dicts back to messages.

    See https://github.com/protocolbuffers/protobuf/issues/2679
    and
    https://developers.google.com/protocol-buffers/docs/proto3#json
    for more details.
    """

    def to_python(self, value, *, absent: bool = None):
        return value

    def to_proto(self, value):
        if value is not None:
            return self._python_type(value)

        return None


class Int64Rule(StringyNumberRule):
    _python_type = int
    _proto_type = ProtoType.INT64


class UInt64Rule(StringyNumberRule):
    _python_type = int
    _proto_type = ProtoType.UINT64


class SInt64Rule(StringyNumberRule):
    _python_type = int
    _proto_type = ProtoType.SINT64


class Fixed64Rule(StringyNumberRule):
    _python_type = int
    _proto_type = ProtoType.FIXED64


class SFixed64Rule(StringyNumberRule):
    _python_type = int
    _proto_type = ProtoType.SFIXED64


STRINGY_NUMBER_RULES = [
    Int64Rule,
    UInt64Rule,
    SInt64Rule,
    Fixed64Rule,
    SFixed64Rule,
]



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


# Underscores may be appended to field names
# that collide with python or proto-plus keywords.
# In case a key only exists with a `_` suffix, coerce the key
# to include the `_` suffix. Is not possible to
# natively define the same field with a trailing underscore in protobuf.
# See related issue
# https://github.com/googleapis/python-api-core/issues/227
def test_fields_mitigate_collision():
    class TestMessage(proto.Message):
        spam_ = proto.Field(proto.STRING, number=1)
        eggs = proto.Field(proto.STRING, number=2)

    obj = TestMessage(spam_="has_spam")
    obj.eggs = "has_eggs"
    assert obj.spam_ == "has_spam"

    # Test that `spam` is coerced to `spam_`
    modified_obj = TestMessage({"spam": "has_spam", "eggs": "has_eggs"})
    assert modified_obj.spam_ == "has_spam"



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock
import inspect
import sys

from google.protobuf import wrappers_pb2

import proto


def test_module_package():
    sys.modules[__name__].__protobuf__ = proto.module(package="spam.eggs.v1")
    try:

        class Foo(proto.Message):
            bar = proto.Field(proto.INT32, number=1)

        marshal = proto.Marshal(name="spam.eggs.v1")

        assert Foo.meta.package == "spam.eggs.v1"
        assert Foo.pb() in marshal._rules
    finally:
        del sys.modules[__name__].__protobuf__


def test_module_package_explicit_marshal():
    sys.modules[__name__].__protobuf__ = proto.module(
        package="spam.eggs.v1", marshal="foo",
    )
    try:

        class Foo(proto.Message):
            bar = proto.Field(proto.INT32, number=1)

        marshal = proto.Marshal(name="foo")

        assert Foo.meta.package == "spam.eggs.v1"
        assert Foo.pb() in marshal._rules
    finally:
        del sys.modules[__name__].__protobuf__


def test_module_manifest():
    __protobuf__ = proto.module(manifest={"Foo", "Bar", "Baz"}, package="spam.eggs.v1",)

    # We want to fake a module, but modules have attribute access, and
    # `frame.f_locals` is a dictionary. Since we only actually care about
    # getattr, this is reasonably easy to shim over.
    frame = inspect.currentframe()
    with mock.patch.object(inspect, "getmodule") as getmodule:
        getmodule.side_effect = lambda *a: View(frame.f_locals)

        class Foo(proto.Message):
            a = proto.Field(wrappers_pb2.Int32Value, number=1)

        class Bar(proto.Message):
            b = proto.Field(proto.MESSAGE, number=1, message=Foo)

        assert not Foo.pb()
        assert not Bar.pb()

        class Baz(proto.Message):
            c = proto.Field(wrappers_pb2.BoolValue, number=1)

        assert Foo.pb()
        assert Bar.pb()
        assert Baz.pb()

        foo = Foo(a=12)
        bar = Bar(b=Foo(a=24))
        baz = Baz(c=False)
        assert foo.a == 12
        assert bar.b.a == 24
        assert baz.c is False


class View:
    """A view around a mapping, for attribute-like access."""

    def __init__(self, mapping):
        self._mapping = mapping

    def __getattr__(self, name):
        if name not in self._mapping:
            raise AttributeError
        return self._mapping[name]



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from proto.utils import cached_property


class MapComposite(collections.abc.MutableMapping):
    """A view around a mutable sequence in protocol buffers.

    This implements the full Python MutableMapping interface, but all methods
    modify the underlying field container directly.
    """

    @cached_property
    def _pb_type(self):
        """Return the protocol buffer type for this sequence."""
        # Huzzah, another hack. Still less bad than RepeatedComposite.
        return type(self.pb.GetEntryClass()().value)

    def __init__(self, sequence, *, marshal):
        """Initialize a wrapper around a protobuf map.

        Args:
            sequence: A protocol buffers map.
            marshal (~.MarshalRegistry): An instantiated marshal, used to
                convert values going to and from this map.
        """
        self._pb = sequence
        self._marshal = marshal

    def __contains__(self, key):
        # Protocol buffers is so permissive that querying for the existence
        # of a key will in of itself create it.
        #
        # By taking a tuple of the keys and querying that, we avoid sending
        # the lookup to protocol buffers and therefore avoid creating the key.
        return key in tuple(self.keys())

    def __getitem__(self, key):
        # We handle raising KeyError ourselves, because otherwise protocol
        # buffers will create the key if it does not exist.
        if key not in self:
            raise KeyError(key)
        return self._marshal.to_python(self._pb_type, self.pb[key])

    def __setitem__(self, key, value):
        pb_value = self._marshal.to_proto(self._pb_type, value, strict=True)

        # Directly setting a key is not allowed; however, protocol buffers
        # is so permissive that querying for the existence of a key will in
        # of itself create it.
        #
        # Therefore, we create a key that way (clearing any fields that may
        # be set) and then merge in our values.
        self.pb[key].Clear()
        self.pb[key].MergeFrom(pb_value)

    def __delitem__(self, key):
        self.pb.pop(key)

    def __len__(self):
        return len(self.pb)

    def __iter__(self):
        return iter(self.pb)

    @property
    def pb(self):
        return self._pb



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import pickle

import pytest

import proto


class Squid(proto.Message):
    # Test primitives, enums, and repeated fields.
    class Chromatophore(proto.Message):
        class Color(proto.Enum):
            UNKNOWN = 0
            RED = 1
            BROWN = 2
            WHITE = 3
            BLUE = 4

        color = proto.Field(Color, number=1)

    mass_kg = proto.Field(proto.INT32, number=1)
    chromatophores = proto.RepeatedField(Chromatophore, number=2)


def test_pickling():

    s = Squid(mass_kg=20)
    colors = ["RED", "BROWN", "WHITE", "BLUE"]
    s.chromatophores = [
        {"color": c} for c in itertools.islice(itertools.cycle(colors), 10)
    ]

    pickled = pickle.dumps(s)

    unpickled = pickle.loads(pickled)

    assert unpickled == s
